/*  display.c
 *
 *  Display functions for xdemorse application
 */

/*
 *  xdemorse: An application to decode Morse code signals to text
 *
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 3 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */


#include "display.h"
#include "shared.h"

/*------------------------------------------------------------------------*/

/*  Display_Detector()
 *
 *  Updates the Detector scope display
 */

  void
Display_Detector( int plot )
{
  /* Points to plot */
  static GdkPoint *points = NULL;
  static int points_idx = 0;
  static cairo_t *cr = NULL;

  /* Initialize on first call */
  if( points == NULL )
  {
	if( !mem_alloc( (void *)&points,
		  (size_t)gbl_wfall_width * sizeof(GdkPoint)) )
	  return;
  }

  /* Initialize cairo */
  if( cr == NULL ) cr = gdk_cairo_create( gbl_scope->window );

  /* Save values to be plotted  */
  points[points_idx].y = gbl_scope_height - plot - 1;
  if( points[points_idx].y <= 0 )
	points[points_idx].y = 1;
  if( points[points_idx].y >= gbl_scope_height )
	points[points_idx].y = gbl_scope_height - 1;

  points[points_idx].x = points_idx;

  /* Recycle buffer idx when full and plot */
  points_idx++;
  if( points_idx == gbl_scope_width )
  {
	int idx;

	/* Draw scope background */
	cairo_set_source_rgb( cr, 0.0, .3, 0.0 );
	cairo_rectangle(
		cr, 0.0, 0.0,
		(double)gbl_scope_width,
		(double)gbl_scope_height );
	cairo_fill( cr );

	/* Plot signal graph */
	cairo_set_source_rgb( cr, 0.0, 1.0, 0.0 );
	cairo_move_to( cr, (double)points[0].x, (double)points[0].y );
	for( idx = 1; idx < points_idx; idx++ )
	  cairo_line_to( cr, (double)points[idx].x, (double)points[idx].y );

	/* Stroke paths */
	cairo_stroke( cr );

	/* Draw Mark/Space threshold references */
	cairo_set_source_rgb( cr, 1.0, 1.0, 1.0 );
	int temp = gbl_scope_height - rc_data.det_threshold;
	cairo_move_to( cr, 0.0, (double)temp );
	cairo_line_to( cr, (double)gbl_scope_width, (double)temp );

	/* Stroke paths */
	cairo_stroke( cr );

	points_idx = 0;
  } /* if( points_idx == gbl_scope_width ) */

} /* Display_Detector() */

/*------------------------------------------------------------------------*/

/*  Display_Signal()
 *
 *  Updates the Signal scope display
 */

  void
Display_Signal( int plot )
{
  /* Points to plot */
  static GdkPoint *points = NULL;
  static int points_idx = 0;
  static cairo_t *cr = NULL;

  /* Initialize on first call */
  if( points == NULL )
  {
	if( !mem_alloc( (void *)&points,
		  (size_t)gbl_wfall_width * sizeof(GdkPoint)) )
	  return;
  }

  /* Initialize cairo */
  if( cr == NULL )
	cr = gdk_cairo_create( gbl_scope->window );

  /* Save values to be plotted (scaled to fit display) */
  points[points_idx].y = gbl_scope_height - plot - 1;
  if( points[points_idx].y <= 0 )
	points[points_idx].y = 1;
  if( points[points_idx].y >= gbl_scope_height )
	points[points_idx].y = gbl_scope_height - 1;

  points[points_idx].x = points_idx;

  /* Recycle buffer idx when full and plot */
  points_idx++;
  if( points_idx == gbl_scope_width )
  {
	int idx;

	/* Draw scope background */
	cairo_set_source_rgb( cr, 0.0, .3, 0.0 );
	cairo_rectangle(
		cr, 0.0, 0.0,
		(double)gbl_scope_width,
		(double)gbl_scope_height );
	cairo_fill( cr );

	/* Plot signal graph */
	cairo_set_source_rgb( cr, 0.0, 1.0, 0.0 );
	cairo_move_to( cr, (double)points[0].x, (double)points[0].y );
	for( idx = 1; idx < points_idx; idx++ )
	  cairo_line_to( cr, (double)points[idx].x, (double)points[idx].y );

	/* Stroke paths */
	cairo_stroke( cr );

	points_idx = 0;
  } /* if( points_idx == DRAWABLE_LEN ) */

} /* Display_Signal */

/*------------------------------------------------------------------------*/

/* DFT_Bin_Value()
 *
 * Calculates DFT bin values with auto level control
 */
  int
DFT_Bin_Value( int sum_i, int sum_q, gboolean reset )
{
 /* Value of dft output "bin" */
  static int bin_val = 0;

  /* Maximum value of dft bins */
  static int bin_max = 1000, max = 0;


  /* Calculate sliding window average of max bin value */
  if( reset )
  {
	bin_max = max;
	if( !bin_max ) bin_max = 1;
	max = 0;
  }
  else
  {
	/* Calculate average signal power at each frequency (bin) */
	bin_val  = bin_val * AMPL_AVE_MUL;
	bin_val += sum_i * sum_i + sum_q * sum_q;
	bin_val /= AMPL_AVE_WIN;

	/* Record max bin value */
	if( max < bin_val ) max = bin_val;

	/* Scale bin values to 255 depending on max value */
	int ret = ( 255 * bin_val ) / bin_max;
	if( ret > 255 ) ret = 255;
	return( ret );
  }

  return( 0 );
} /* DFT_Bin_Value() */

/*------------------------------------------------------------------------*/

/* Display_Waterfall()
 *
 * Displays audio spectrum as "waterfall"
 */

  void
Display_Waterfall( void )
{
  int
	idh, idv,  /* Index to hor. and vert. position in warterfall */
	pixel_val, /* Greyscale value of pixel derived from dft o/p  */
	dft_idx;   /* Index to dft output array */

  /* Constant needed to draw center line in waterfall */
  int center_line = gbl_wfall_width / 2;

  /* Pointer to current pixel */
  static guchar *pix;

  /* Draw a vertical white line in waterfall at detector's freq. */
  if( isFlagSet(CENTERLINE) )
  {
	pix = gbl_wfall_pixels + gbl_wfall_rowstride + gbl_wfall_n_channels;
	pix += gbl_wfall_n_channels * center_line;
	pix[0] = pix[1] = pix[2] = 0xff;
  }

  /* Copy each line of waterfall to next one */
  for( idv = gbl_wfall_height-2; idv > 0; idv-- )
  {
	pix = gbl_wfall_pixels + gbl_wfall_rowstride * idv + gbl_wfall_n_channels;

	for( idh = 0; idh < gbl_wfall_width; idh++ )
	{
	  pix[0] = pix[ -gbl_wfall_rowstride];
	  pix[1] = pix[1-gbl_wfall_rowstride];
	  pix[2] = pix[2-gbl_wfall_rowstride];
	  pix += gbl_wfall_n_channels;
	}
  }

  /* Go to top left +1 hor. +1 vert. of pixbuf */
  pix = gbl_wfall_pixels + gbl_wfall_rowstride + gbl_wfall_n_channels;

  /* First column of display (DC) not used */
  gbl_bin_ave[0] = 0;

  /* Do the DFT on input array */
  Idft( DFT_INPUT_SIZE, gbl_wfall_width );
  for( dft_idx = 0; dft_idx < gbl_wfall_width; dft_idx++ )
  {
	/* Calculate signal power at each frequency (bin) */
	dft_out_r[dft_idx] /= 1024;
	dft_out_i[dft_idx] /= 1024;
	pixel_val = DFT_Bin_Value(
		dft_out_r[dft_idx], dft_out_i[dft_idx], FALSE );
	gbl_bin_ave[dft_idx] = pixel_val;

	/* Color code signal strength */
	int n;
	if( pixel_val < 85 )
	{
	  pix[0] = 0;
	  pix[1] = 0;
	  pix[2] = 3 + (guchar)pixel_val * 3;
	}
	else if( pixel_val < 170 )
	{
	  n = pixel_val - 85;
	  pix[0] = 0;
	  pix[1] = 3 + (guchar)n * 3;
	  pix[2] = 255 - (guchar)n * 3;
	}
	else
	{
	  n = pixel_val - 170;
	  pix[0] = (guchar)pixel_val;
	  pix[1] = 255;
	  pix[2] = 0;
	}

	pix += gbl_wfall_n_channels;

  } /* for( dft_idx = 0; dft_idx < dft_end; dft_idx++ ) */

  /* Reset function */
  pixel_val = DFT_Bin_Value( 0, 0, TRUE );

  /* At last draw waterfall */
  gtk_widget_queue_draw( gbl_waterfall );

} /* Display_Waterfall() */

/*------------------------------------------------------------------------*/

